@@ -6,12 +6,14 @@ use holo_hash::hash_type::AnyDht;
 use holo_hash::AnyDhtHash;
 use holo_hash::EntryHash;
 use holo_hash::HeaderHash;
-use holochain_types::prelude::Judged;
+use holochain_keystore::KeystoreError;
+use holochain_types::prelude::*;
 use holochain_zome_types::entry::EntryHashed;
 use holochain_zome_types::ChainTopOrdering;
 use holochain_zome_types::Element;
 use holochain_zome_types::Entry;
 use holochain_zome_types::SignedHeaderHashed;
+use holochain_zome_types::TimestampError;
 use thiserror::Error;
 
 use crate::prelude::Query;
@@ -169,6 +171,7 @@ impl SyncScratch {
             .lock()
             .map_err(|_| SyncScratchError::ScratchLockPoison)?))
     }
+
     pub fn apply_and_then<T, E, F>(&self, f: F) -> Result<T, E>
     where
         E: From<SyncScratchError>,
@@ -239,6 +242,16 @@ impl StoresIter<Judged<SignedHeaderHashed>> for FilteredScratch {
     }
 }
 
+#[derive(Error, Debug)]
+pub enum ScratchError {
+    #[error(transparent)]
+    Timestamp(#[from] TimestampError),
+    #[error(transparent)]
+    Keystore(#[from] KeystoreError),
+    #[error(transparent)]
+    Header(#[from] HeaderError),
+}
+
 #[derive(Error, Debug)]
 pub enum SyncScratchError {
     #[error("Scratch lock was poisoned")]
